@@ -118,6 +118,7 @@ protected:
 
     ComplexVector ritz_val; // ritz values
     ComplexMatrix ritz_vec; // ritz vectors
+    ComplexVector ritz_est; // last row of ritz_vec
 
 private:
     BoolArray ritz_conv;    // indicator of the convergence of ritz values
@@ -135,17 +136,17 @@ private:
         fac_f = fk;
 
         Vector w(dim_n);
-        Scalar beta = 0.0;
+        Scalar beta = fac_f.norm();
         // Keep the upperleft k x k submatrix of H and set other elements to 0
         fac_H.rightCols(ncv - from_k).setZero();
         fac_H.block(from_k, 0, ncv - from_k, from_k).setZero();
         for(int i = from_k; i <= to_m - 1; i++)
         {
-            beta = fac_f.norm();
+            // v <- f / ||f||
             fac_V.col(i).noalias() = fac_f / beta; // The (i+1)-th column
             fac_H(i, i - 1) = beta;
 
-            // w = A * v, v = fac_V.col(i)
+            // w <- A * v, v = fac_V.col(i)
             op->perform_op(&fac_V(0, i), w.data());
             nmatop++;
 
@@ -153,20 +154,21 @@ private:
             MapMat Vs(fac_V.data(), dim_n, i + 1);
             // h = fac_H(0:i, i)
             MapVec h(&fac_H(0, i), i + 1);
+            // h <- V' * w
             h.noalias() = Vs.transpose() * w;
 
-            fac_f = w - Vs * h;
-            // Correct f if it is not orthogonal to V
-            // Typically the largest absolute value occurs in
-            // the first element, i.e., <v1, f>, so we use this
-            // to test the orthogonality
-            Scalar v1f = fac_f.dot(fac_V.col(0));
-            if(v1f > prec || v1f < -prec)
+            // f <- w - V * h
+            fac_f.noalias() = w - Vs * h;
+            beta = fac_f.norm();
+
+            // f/||f|| is going to be the next column of V, so we need to test
+            // whether V' * (f/||f||) ~= 0
+            Vector Vf = Vs.transpose() * fac_f;
+            if(Vf.cwiseAbs().maxCoeff() > prec * beta)
             {
-                Vector Vf(i + 1);
-                Vf.tail(i) = fac_V.block(0, 1, dim_n, i).transpose() * fac_f;
-                Vf[0] = v1f;
-                fac_f -= Vs * Vf;
+                // f <- f - V * V' * f
+                fac_f.noalias() -= Vs * Vf;
+                beta = fac_f.norm();
             }
         }
     }
@@ -252,7 +254,7 @@ private:
     {
         // thresh = tol * max(prec, abs(theta)), theta for ritz value
         Array thresh = tol * ritz_val.head(nev).array().abs().max(prec);
-        Array resid = ritz_vec.template bottomRows<1>().transpose().array().abs() * fac_f.norm();
+        Array resid = ritz_est.head(nev).array().abs() * fac_f.norm();
         // Converged "wanted" ritz values
         ritz_conv = (resid < thresh);
 
@@ -264,15 +266,18 @@ private:
     {
         int nev_new = nev;
 
+        for(int i = nev; i < ncv; i++)
+            if(std::abs(ritz_est[i]) < prec)  nev_new++;
+
         // Increase nev by one if ritz_val[nev - 1] and
         // ritz_val[nev] are conjugate pairs
-        if(is_complex(ritz_val[nev - 1], prec) &&
-           is_conj(ritz_val[nev - 1], ritz_val[nev], prec))
+        if(is_complex(ritz_val[nev_new - 1], prec) &&
+           is_conj(ritz_val[nev_new - 1], ritz_val[nev_new], prec))
         {
-            nev_new = nev + 1;
+            nev_new++;
         }
         // Adjust nev_new again, according to dnaup2.f line 660~674 in ARPACK
-        nev_new = nev_new + std::min(nconv, (ncv - nev_new) / 2);
+        nev_new += std::min(nconv, (ncv - nev_new) / 2);
         if(nev_new == 1 && ncv >= 6)
             nev_new = ncv / 2;
         else if(nev_new == 1 && ncv > 3)
@@ -305,6 +310,7 @@ private:
         for(int i = 0; i < ncv; i++)
         {
             ritz_val[i] = evals[ind[i]];
+            ritz_est[i] = evecs(ncv - 1, ind[i]);
         }
         for(int i = 0; i < nev; i++)
         {
@@ -426,6 +432,7 @@ public:
         fac_f.resize(dim_n);
         ritz_val.resize(ncv);
         ritz_vec.resize(ncv, nev);
+        ritz_est.resize(ncv);
         ritz_conv.resize(nev);
 
         fac_V.setZero();
@@ -433,6 +440,7 @@ public:
         fac_f.setZero();
         ritz_val.setZero();
         ritz_vec.setZero();
+        ritz_est.setZero();
         ritz_conv.setZero();
 
         Vector v(dim_n);
