@@ -183,6 +183,7 @@ protected:
 
 private:
     Matrix ritz_vec;     // ritz vectors
+    Vector ritz_est;     // last row of ritz_vec
     BoolArray ritz_conv; // indicator of the convergence of ritz values
 
     const Scalar prec;   // precision parameter used to test convergence
@@ -198,17 +199,18 @@ private:
         fac_f = fk;
 
         Vector w(dim_n);
-        Scalar beta = 0.0, Hii = 0.0;
+        Scalar beta = fac_f.norm(), Hii = 0.0;
         // Keep the upperleft k x k submatrix of H and set other elements to 0
         fac_H.rightCols(ncv - from_k).setZero();
         fac_H.block(from_k, 0, ncv - from_k, from_k).setZero();
         for(int i = from_k; i <= to_m - 1; i++)
         {
-            beta = fac_f.norm();
+            // v <- f / ||f||
             MapVec v(&fac_V(0, i), dim_n); // The (i+1)-th column
             v.noalias() = fac_f / beta;
             fac_H(i, i - 1) = beta;
 
+            // w <- A * v
             op->perform_op(v.data(), w.data());
             nmatop++;
 
@@ -216,18 +218,19 @@ private:
             fac_H(i - 1, i) = beta;
             fac_H(i, i) = Hii;
 
+            //  f <- w - V * V' * w
             fac_f.noalias() = w - beta * fac_V.col(i - 1) - Hii * v;
-            // Correct f if it is not orthogonal to V
-            // Typically the largest absolute value occurs in
-            // the first element, i.e., <v1, f>, so we use this
-            // to test the orthogonality
-            Scalar v1f = fac_f.dot(fac_V.col(0));
-            if(v1f > prec || v1f < -prec)
+            beta = fac_f.norm();
+
+            // f/||f|| is going to be the next column of V, so we need to test
+            // whether V' * (f/||f||) ~= 0
+            MapMat V(fac_V.data(), dim_n, i + 1); // The first (i+1) columns
+            Vector Vf = V.transpose() * fac_f;
+            if(Vf.cwiseAbs().maxCoeff() > prec * beta)
             {
-                Vector Vf(i + 1);
-                Vf.tail(i) = fac_V.block(0, 1, dim_n, i).transpose() * fac_f;
-                Vf[0] = v1f;
-                fac_f -= fac_V.leftCols(i + 1) * Vf;
+                // f <- f - V * V' * f
+                fac_f.noalias() -= V * Vf;
+                beta = fac_f.norm();
             }
         }
     }
@@ -280,7 +283,7 @@ private:
     {
         // thresh = tol * max(prec, abs(theta)), theta for ritz value
         Array thresh = tol * ritz_val.head(nev).array().abs().max(prec);
-        Array resid =  ritz_vec.template bottomRows<1>().transpose().array().abs() * fac_f.norm();
+        Array resid =  ritz_est.head(nev).array().abs() * fac_f.norm();
         // Converged "wanted" ritz values
         ritz_conv = (resid < thresh);
 
@@ -292,11 +295,14 @@ private:
     {
         int nev_new = nev;
 
+        for(int i = nev; i < ncv; i++)
+            if(std::abs(ritz_est[i]) < prec)  nev_new++;
+
         // Adjust nev_new, according to dsaup2.f line 677~684 in ARPACK
-        nev_new = nev + std::min(nconv, (ncv - nev) / 2);
-        if(nev == 1 && ncv >= 6)
+        nev_new += std::min(nconv, (ncv - nev_new) / 2);
+        if(nev_new == 1 && ncv >= 6)
             nev_new = ncv / 2;
-        else if(nev == 1 && ncv > 2)
+        else if(nev_new == 1 && ncv > 2)
             nev_new = 2;
 
         return nev_new;
@@ -338,6 +344,7 @@ private:
         for(int i = 0; i < ncv; i++)
         {
             ritz_val[i] = evals[ind[i]];
+            ritz_est[i] = evecs(ncv - 1, ind[i]);
         }
         for(int i = 0; i < nev; i++)
         {
@@ -447,6 +454,7 @@ public:
         fac_f.resize(dim_n);
         ritz_val.resize(ncv);
         ritz_vec.resize(ncv, nev);
+        ritz_est.resize(ncv);
         ritz_conv.resize(nev);
 
         fac_V.setZero();
@@ -454,6 +462,7 @@ public:
         fac_f.setZero();
         ritz_val.setZero();
         ritz_vec.setZero();
+        ritz_est.setZero();
         ritz_conv.setZero();
 
         nmatop = 0;
